@@ -8,9 +8,8 @@ use crate::memory::{Memory, RuntimeMemoryCreator};
 use crate::table::{Table, TableElement};
 use crate::traphandlers::Trap;
 use crate::vmcontext::{
-    VMBuiltinFunctionsArray, VMCallerCheckedAnyfunc, VMContext, VMFunctionImport,
-    VMGlobalDefinition, VMGlobalImport, VMInterrupts, VMMemoryDefinition, VMMemoryImport,
-    VMSharedSignatureIndex, VMTableDefinition, VMTableImport,
+    VMCallerCheckedAnyfunc, VMContext, VMFunctionImport, VMGlobalDefinition, VMGlobalImport,
+    VMInterrupts, VMMemoryDefinition, VMMemoryImport, VMTableDefinition, VMTableImport,
 };
 use crate::{ExportFunction, ExportGlobal, ExportMemory, ExportTable, Store};
 use memoffset::offset_of;
@@ -28,7 +27,7 @@ use wasmtime_environ::wasm::{
     DataIndex, DefinedGlobalIndex, DefinedMemoryIndex, DefinedTableIndex, ElemIndex, EntityIndex,
     FuncIndex, GlobalIndex, MemoryIndex, TableElementType, TableIndex, WasmType,
 };
-use wasmtime_environ::{ir, Module, VMOffsets};
+use wasmtime_environ::{ir, HostPtr, Module, VMOffsets};
 
 mod allocator;
 
@@ -99,21 +98,39 @@ pub trait ResourceLimiter {
     }
 }
 
-/// A WebAssembly instance.
+/// A type that roughly corresponds to a WebAssembly instance, but is also used
+/// for host-defined objects.
 ///
-/// This is repr(C) to ensure that the vmctx field is last.
-#[repr(C)]
+/// This structure is is never allocated directly but is instead managed through
+/// an `InstanceHandle`. This structure ends with a `VMContext` which has a
+/// dynamic size corresponding to the `module` configured within. Memory
+/// management of this structure is always externalized.
+///
+/// Instances here can correspond to actual instantiated modules, but it's also
+/// used ubiquitously for host-defined objects. For example creating a
+/// host-defined memory will have a `module` that looks like it exports a single
+/// memory (and similar for other constructs).
+///
+/// This `Instance` type is used as a ubiquitous representation for WebAssembly
+/// values, whether or not they were created on the host or through a module.
+#[repr(C)] // ensure that the vmctx field is last.
 pub(crate) struct Instance {
     /// The `Module` this `Instance` was instantiated from.
     module: Arc<Module>,
 
-    /// Offsets in the `vmctx` region.
-    offsets: VMOffsets,
+    /// Offsets in the `vmctx` region, precomputed from the `module` above.
+    offsets: VMOffsets<HostPtr>,
 
     /// WebAssembly linear memory data.
+    ///
+    /// This is where all runtime information about defined linear memories in
+    /// this module lives.
     memories: PrimaryMap<DefinedMemoryIndex, Memory>,
 
     /// WebAssembly table data.
+    ///
+    /// Like memories, this is only for defined tables in the module and
+    /// contains all of their runtime state.
     tables: PrimaryMap<DefinedTableIndex, Table>,
 
     /// Stores the dropped passive element segments in this instantiation by index.
@@ -125,6 +142,9 @@ pub(crate) struct Instance {
     dropped_data: EntitySet<DataIndex>,
 
     /// Hosts can store arbitrary per-instance information here.
+    ///
+    /// Most of the time from Wasmtime this is `Box::new(())`, a noop
+    /// allocation, but some host-defined objects will store their state here.
     host_state: Box<dyn Any + Send + Sync>,
 
     /// Additional context used by compiled wasm code. This field is last, and
@@ -143,57 +163,28 @@ impl Instance {
             .cast()
     }
 
-    pub(crate) fn module(&self) -> &Module {
+    pub(crate) fn module(&self) -> &Arc<Module> {
         &self.module
     }
 
-    /// Return a pointer to the `VMSharedSignatureIndex`s.
-    fn signature_ids_ptr(&self) -> *mut VMSharedSignatureIndex {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_signature_ids_begin()) }
-    }
-
     /// Return the indexed `VMFunctionImport`.
     fn imported_function(&self, index: FuncIndex) -> &VMFunctionImport {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { &*self.imported_functions_ptr().add(index) }
-    }
-
-    /// Return a pointer to the `VMFunctionImport`s.
-    fn imported_functions_ptr(&self) -> *mut VMFunctionImport {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_imported_functions_begin()) }
+        unsafe { &*self.vmctx_plus_offset(self.offsets.vmctx_vmfunction_import(index)) }
     }
 
     /// Return the index `VMTableImport`.
     fn imported_table(&self, index: TableIndex) -> &VMTableImport {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { &*self.imported_tables_ptr().add(index) }
-    }
-
-    /// Return a pointer to the `VMTableImports`s.
-    fn imported_tables_ptr(&self) -> *mut VMTableImport {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_imported_tables_begin()) }
+        unsafe { &*self.vmctx_plus_offset(self.offsets.vmctx_vmtable_import(index)) }
     }
 
     /// Return the indexed `VMMemoryImport`.
     fn imported_memory(&self, index: MemoryIndex) -> &VMMemoryImport {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { &*self.imported_memories_ptr().add(index) }
-    }
-
-    /// Return a pointer to the `VMMemoryImport`s.
-    fn imported_memories_ptr(&self) -> *mut VMMemoryImport {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_imported_memories_begin()) }
+        unsafe { &*self.vmctx_plus_offset(self.offsets.vmctx_vmmemory_import(index)) }
     }
 
     /// Return the indexed `VMGlobalImport`.
     fn imported_global(&self, index: GlobalIndex) -> &VMGlobalImport {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { &*self.imported_globals_ptr().add(index) }
-    }
-
-    /// Return a pointer to the `VMGlobalImport`s.
-    fn imported_globals_ptr(&self) -> *mut VMGlobalImport {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_imported_globals_begin()) }
+        unsafe { &*self.vmctx_plus_offset(self.offsets.vmctx_vmglobal_import(index)) }
     }
 
     /// Return the indexed `VMTableDefinition`.
@@ -211,13 +202,7 @@ impl Instance {
 
     /// Return the indexed `VMTableDefinition`.
     fn table_ptr(&self, index: DefinedTableIndex) -> *mut VMTableDefinition {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { self.tables_ptr().add(index) }
-    }
-
-    /// Return a pointer to the `VMTableDefinition`s.
-    fn tables_ptr(&self) -> *mut VMTableDefinition {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_tables_begin()) }
+        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_vmtable_definition(index)) }
     }
 
     /// Get a locally defined or imported memory.
@@ -244,13 +229,7 @@ impl Instance {
 
     /// Return the indexed `VMMemoryDefinition`.
     fn memory_ptr(&self, index: DefinedMemoryIndex) -> *mut VMMemoryDefinition {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { self.memories_ptr().add(index) }
-    }
-
-    /// Return a pointer to the `VMMemoryDefinition`s.
-    fn memories_ptr(&self) -> *mut VMMemoryDefinition {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_memories_begin()) }
+        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_vmmemory_definition(index)) }
     }
 
     /// Return the indexed `VMGlobalDefinition`.
@@ -260,8 +239,7 @@ impl Instance {
 
     /// Return the indexed `VMGlobalDefinition`.
     fn global_ptr(&self, index: DefinedGlobalIndex) -> *mut VMGlobalDefinition {
-        let index = usize::try_from(index.as_u32()).unwrap();
-        unsafe { self.globals_ptr().add(index) }
+        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_vmglobal_definition(index)) }
     }
 
     /// Get a raw pointer to the global at the given index regardless whether it
@@ -279,16 +257,6 @@ impl Instance {
         }
     }
 
-    /// Return a pointer to the `VMGlobalDefinition`s.
-    fn globals_ptr(&self) -> *mut VMGlobalDefinition {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_globals_begin()) }
-    }
-
-    /// Return a pointer to the `VMBuiltinFunctionsArray`.
-    fn builtin_functions_ptr(&self) -> *mut VMBuiltinFunctionsArray {
-        unsafe { self.vmctx_plus_offset(self.offsets.vmctx_builtin_functions_begin()) }
-    }
-
     /// Return a pointer to the interrupts structure
     pub fn interrupts(&self) -> *mut *const VMInterrupts {
         unsafe { self.vmctx_plus_offset(self.offsets.vmctx_interrupts()) }
@@ -410,32 +378,26 @@ impl Instance {
     }
 
     /// Return the table index for the given `VMTableDefinition`.
-    pub(crate) fn table_index(&self, table: &VMTableDefinition) -> DefinedTableIndex {
-        let offsets = &self.offsets;
-        let begin = unsafe {
-            (&self.vmctx as *const VMContext as *const u8)
-                .add(usize::try_from(offsets.vmctx_tables_begin()).unwrap())
-        } as *const VMTableDefinition;
-        let end: *const VMTableDefinition = table;
-        // TODO: Use `offset_from` once it stablizes.
+    unsafe fn table_index(&self, table: &VMTableDefinition) -> DefinedTableIndex {
         let index = DefinedTableIndex::new(
-            (end as usize - begin as usize) / mem::size_of::<VMTableDefinition>(),
+            usize::try_from(
+                (table as *const VMTableDefinition)
+                    .offset_from(self.table_ptr(DefinedTableIndex::new(0))),
+            )
+            .unwrap(),
         );
         assert_lt!(index.index(), self.tables.len());
         index
     }
 
     /// Return the memory index for the given `VMMemoryDefinition`.
-    pub(crate) fn memory_index(&self, memory: &VMMemoryDefinition) -> DefinedMemoryIndex {
-        let offsets = &self.offsets;
-        let begin = unsafe {
-            (&self.vmctx as *const VMContext as *const u8)
-                .add(usize::try_from(offsets.vmctx_memories_begin()).unwrap())
-        } as *const VMMemoryDefinition;
-        let end: *const VMMemoryDefinition = memory;
-        // TODO: Use `offset_from` once it stablizes.
+    unsafe fn memory_index(&self, memory: &VMMemoryDefinition) -> DefinedMemoryIndex {
         let index = DefinedMemoryIndex::new(
-            (end as usize - begin as usize) / mem::size_of::<VMMemoryDefinition>(),
+            usize::try_from(
+                (memory as *const VMMemoryDefinition)
+                    .offset_from(self.memory_ptr(DefinedMemoryIndex::new(0))),
+            )
+            .unwrap(),
         );
         assert_lt!(index.index(), self.memories.len());
         index
@@ -445,68 +407,31 @@ impl Instance {
     ///
     /// Returns `None` if memory can't be grown by the specified amount
     /// of pages.
-    pub(crate) fn memory_grow(
-        &mut self,
-        memory_index: DefinedMemoryIndex,
-        delta: u32,
-    ) -> Option<u32> {
-        let limiter = unsafe { (*self.store()).limiter() };
-        let memory = self
-            .memories
-            .get_mut(memory_index)
-            .unwrap_or_else(|| panic!("no memory for index {}", memory_index.index()));
+    pub(crate) fn memory_grow(&mut self, index: MemoryIndex, delta: u32) -> Option<u32> {
+        let (idx, instance) = if let Some(idx) = self.module.defined_memory_index(index) {
+            (idx, self)
+        } else {
+            let import = self.imported_memory(index);
+            unsafe {
+                let foreign_instance = (*import.vmctx).instance_mut();
+                let foreign_memory_def = &*import.from;
+                let foreign_memory_index = foreign_instance.memory_index(foreign_memory_def);
+                (foreign_memory_index, foreign_instance)
+            }
+        };
+        let limiter = unsafe { (*instance.store()).limiter() };
+        let memory = &mut instance.memories[idx];
 
         let result = unsafe { memory.grow(delta, limiter) };
+        let vmmemory = memory.vmmemory();
 
-        // Keep current the VMContext pointers used by compiled wasm code.
-        self.set_memory(memory_index, self.memories[memory_index].vmmemory());
+        // Update the state used by wasm code in case the base pointer and/or
+        // the length changed.
+        instance.set_memory(idx, vmmemory);
 
         result
     }
 
-    /// Grow imported memory by the specified amount of pages.
-    ///
-    /// Returns `None` if memory can't be grown by the specified amount
-    /// of pages.
-    ///
-    /// # Safety
-    /// This and `imported_memory_size` are currently unsafe because they
-    /// dereference the memory import's pointers.
-    pub(crate) unsafe fn imported_memory_grow(
-        &mut self,
-        memory_index: MemoryIndex,
-        delta: u32,
-    ) -> Option<u32> {
-        let import = self.imported_memory(memory_index);
-        let foreign_instance = (*import.vmctx).instance_mut();
-        let foreign_memory = &*import.from;
-        let foreign_index = foreign_instance.memory_index(foreign_memory);
-
-        foreign_instance.memory_grow(foreign_index, delta)
-    }
-
-    /// Returns the number of allocated wasm pages.
-    pub(crate) fn memory_size(&self, memory_index: DefinedMemoryIndex) -> u32 {
-        self.memories
-            .get(memory_index)
-            .unwrap_or_else(|| panic!("no memory for index {}", memory_index.index()))
-            .size()
-    }
-
-    /// Returns the number of allocated wasm pages in an imported memory.
-    ///
-    /// # Safety
-    /// This and `imported_memory_grow` are currently unsafe because they
-    /// dereference the memory import's pointers.
-    pub(crate) unsafe fn imported_memory_size(&self, memory_index: MemoryIndex) -> u32 {
-        let import = self.imported_memory(memory_index);
-        let foreign_instance = (&mut *import.vmctx).instance();
-        let foreign_memory = &mut *import.from;
-        let foreign_index = foreign_instance.memory_index(foreign_memory);
-
-        foreign_instance.memory_size(foreign_index)
-    }
-
     pub(crate) fn table_element_type(&mut self, table_index: TableIndex) -> TableElementType {
         unsafe { (*self.get_table(table_index)).element_type() }
     }
@@ -701,13 +626,11 @@ impl Instance {
         let src_mem = self.get_memory(src_index);
         let dst_mem = self.get_memory(dst_index);
 
-        if src
-            .checked_add(len)
-            .map_or(true, |n| n as usize > src_mem.current_length)
-            || dst
-                .checked_add(len)
-                .map_or(true, |m| m as usize > dst_mem.current_length)
-        {
+        if src.checked_add(len).map_or(true, |n| {
+            usize::try_from(n).unwrap() > src_mem.current_length
+        }) || dst.checked_add(len).map_or(true, |m| {
+            usize::try_from(m).unwrap() > dst_mem.current_length
+        }) {
             return Err(Trap::wasm(ir::TrapCode::HeapOutOfBounds));
         }
 
@@ -730,19 +653,18 @@ impl Instance {
     /// # Errors
     ///
     /// Returns a `Trap` error if the memory range is out of bounds.
-    pub(crate) fn defined_memory_fill(
-        &self,
-        memory_index: DefinedMemoryIndex,
+    pub(crate) fn memory_fill(
+        &mut self,
+        memory_index: MemoryIndex,
         dst: u32,
         val: u32,
         len: u32,
     ) -> Result<(), Trap> {
-        let memory = self.memory(memory_index);
+        let memory = self.get_memory(memory_index);
 
-        if dst
-            .checked_add(len)
-            .map_or(true, |m| m as usize > memory.current_length)
-        {
+        if dst.checked_add(len).map_or(true, |m| {
+            usize::try_from(m).unwrap() > memory.current_length
+        }) {
             return Err(Trap::wasm(ir::TrapCode::HeapOutOfBounds));
         }
 
@@ -759,27 +681,6 @@ impl Instance {
         Ok(())
     }
 
-    /// Perform the `memory.fill` operation on an imported memory.
-    ///
-    /// # Errors
-    ///
-    /// Returns a `Trap` error if the memory range is out of bounds.
-    pub(crate) fn imported_memory_fill(
-        &self,
-        memory_index: MemoryIndex,
-        dst: u32,
-        val: u32,
-        len: u32,
-    ) -> Result<(), Trap> {
-        let import = self.imported_memory(memory_index);
-        unsafe {
-            let foreign_instance = (&*import.vmctx).instance();
-            let foreign_memory = &*import.from;
-            let foreign_index = foreign_instance.memory_index(foreign_memory);
-            foreign_instance.defined_memory_fill(foreign_index, dst, val, len)
-        }
-    }
-
     /// Performs the `memory.init` operation.
     ///
     /// # Errors
@@ -822,10 +723,10 @@ impl Instance {
 
         if src
             .checked_add(len)
-            .map_or(true, |n| n as usize > data.len())
-            || dst
-                .checked_add(len)
-                .map_or(true, |m| m as usize > memory.current_length)
+            .map_or(true, |n| usize::try_from(n).unwrap() > data.len())
+            || dst.checked_add(len).map_or(true, |m| {
+                usize::try_from(m).unwrap() > memory.current_length
+            })
         {
             return Err(Trap::wasm(ir::TrapCode::HeapOutOfBounds));
         }
@@ -854,11 +755,8 @@ impl Instance {
     /// Get a table by index regardless of whether it is locally-defined or an
     /// imported, foreign table.
     pub(crate) fn get_table(&mut self, table_index: TableIndex) -> *mut Table {
-        if let Some(defined_table_index) = self.module.defined_table_index(table_index) {
-            self.get_defined_table(defined_table_index)
-        } else {
-            self.get_foreign_table(table_index)
-        }
+        let (idx, instance) = self.get_defined_table_index_and_instance(table_index);
+        ptr::addr_of_mut!(instance.tables[idx])
     }
 
     /// Get a locally-defined table.
@@ -866,15 +764,6 @@ impl Instance {
         ptr::addr_of_mut!(self.tables[index])
     }
 
-    /// Get an imported, foreign table.
-    pub(crate) fn get_foreign_table(&mut self, index: TableIndex) -> *mut Table {
-        let import = self.imported_table(index);
-        let foreign_instance = unsafe { (*import.vmctx).instance_mut() };
-        let foreign_table = unsafe { &*import.from };
-        let foreign_index = foreign_instance.table_index(foreign_table);
-        ptr::addr_of_mut!(foreign_instance.tables[foreign_index])
-    }
-
     pub(crate) fn get_defined_table_index_and_instance(
         &mut self,
         index: TableIndex,
@@ -883,10 +772,12 @@ impl Instance {
             (defined_table_index, self)
         } else {
             let import = self.imported_table(index);
-            let foreign_instance = unsafe { (*import.vmctx).instance_mut() };
-            let foreign_table_def = unsafe { &*import.from };
-            let foreign_table_index = foreign_instance.table_index(foreign_table_def);
-            (foreign_table_index, foreign_instance)
+            unsafe {
+                let foreign_instance = (*import.vmctx).instance_mut();
+                let foreign_table_def = &*import.from;
+                let foreign_table_index = foreign_instance.table_index(foreign_table_def);
+                (foreign_table_index, foreign_instance)
+            }
         }
     }
 
@@ -957,7 +848,7 @@ impl InstanceHandle {
     }
 
     /// Return a reference to a module.
-    pub fn module(&self) -> &Module {
+    pub fn module(&self) -> &Arc<Module> {
         self.instance().module()
     }
 
@@ -981,7 +872,7 @@ impl InstanceHandle {
     }
 
     /// Return the memory index for the given `VMMemoryDefinition` in this instance.
-    pub fn memory_index(&self, memory: &VMMemoryDefinition) -> DefinedMemoryIndex {
+    pub unsafe fn memory_index(&self, memory: &VMMemoryDefinition) -> DefinedMemoryIndex {
         self.instance().memory_index(memory)
     }
 
@@ -991,7 +882,7 @@ impl InstanceHandle {
     }
 
     /// Return the table index for the given `VMTableDefinition` in this instance.
-    pub fn table_index(&self, table: &VMTableDefinition) -> DefinedTableIndex {
+    pub unsafe fn table_index(&self, table: &VMTableDefinition) -> DefinedTableIndex {
         self.instance().table_index(table)
     }
 
