@@ -33,7 +33,7 @@
 
 #![allow(non_snake_case)]
 
-use crate::traphandlers::{tls, Trap, Unwind};
+use crate::traphandlers::{tls, wasmtime_longjmp, Trap};
 use mach::exception_types::*;
 use mach::kern_return::*;
 use mach::mach_init::*;
@@ -389,21 +389,7 @@ unsafe extern "C" fn unwind(wasm_pc: *const u8) -> ! {
         state.jmp_buf.get()
     });
     debug_assert!(!jmp_buf.is_null());
-    Unwind(jmp_buf);
-}
-
-thread_local! {
-    static MY_PORT: ClosePort = ClosePort(unsafe { mach_thread_self() });
-}
-
-struct ClosePort(mach_port_name_t);
-
-impl Drop for ClosePort {
-    fn drop(&mut self) {
-        unsafe {
-            mach_port_deallocate(mach_task_self(), self.0);
-        }
-    }
+    wasmtime_longjmp(jmp_buf);
 }
 
 /// Exceptions on macOS can be delivered to either thread-level or task-level
@@ -426,13 +412,15 @@ impl Drop for ClosePort {
 pub fn lazy_per_thread_init() -> Result<(), Trap> {
     unsafe {
         assert!(WASMTIME_PORT != MACH_PORT_NULL);
+        let this_thread = mach_thread_self();
         let kret = thread_set_exception_ports(
-            MY_PORT.with(|p| p.0),
+            this_thread,
             EXC_MASK_BAD_ACCESS | EXC_MASK_BAD_INSTRUCTION,
             WASMTIME_PORT,
             EXCEPTION_DEFAULT | MACH_EXCEPTION_CODES,
             mach_addons::THREAD_STATE_NONE,
         );
+        mach_port_deallocate(mach_task_self(), this_thread);
         assert_eq!(kret, KERN_SUCCESS, "failed to set thread exception port");
     }
     Ok(())
